#include "linuxmodver.h"

#include "KThread.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#else
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/errno.h>

#include <asm/unistd.h>
#endif

#include <linux/smp_lock.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static void kthread_launcher(
    void *data)
{
  kthread_t *kthread = data;

  kthread->pid = kernel_thread((int (*)(void *))kthread->function, kthread, 0);
}
#endif

/* create a new kernel thread. Called by the creator. */
void start_kthread(
    KTHREAD_R (*func)(kthread_t *),
    kthread_t *kthread)
{
  /* initialize the semaphore:
     we start with the semaphore locked. The new kernel
     thread will setup its stuff and unlock it. This
     control flow (the one that creates the thread) blocks
     in the down operation below until the thread has reached
     the up() operation.
  */
  init_completion(&kthread->startstop_sem);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  /* store the function to be executed in the data passed to
     the launcher */
  kthread->function = func;

  /* create the new thread by running a task through keventd */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  INIT_WORK(&kthread->tq, kthread_launcher, kthread);

  /* and schedule it for execution */
  schedule_work(&kthread->tq);
#else
  /* initialize the task queue structure */
  kthread->tq.sync = 0;
  INIT_LIST_HEAD(&kthread->tq.list);
  kthread->tq.routine = kthread_launcher;
  kthread->tq.data = kthread;

  /* and schedule it for execution */
  schedule_task(&kthread->tq);
#endif
#else
  kthread->pid = kernel_thread(func, kthread, 0);
#endif

  /* wait till it has reached the setup_thread routine */
  wait_for_completion(&kthread->startstop_sem);
}

/* stop a kernel thread. Called by the removing instance */
void stop_kthread(kthread_t *kthread)
{
  if (kthread->pid <= 0)
  {
    printk("stop_kthread: killing non existing thread!\n");
    return;
  }

  /* this function needs to be protected with the big
     kernel lock (lock_kernel()). The lock must be
     grabbed before changing the terminate
     flag and released after the down() call. */
#ifdef CONFIG_SMP
  lock_kernel();
#endif

  /* initialize the semaphore. We lock it here, the
     leave_thread call of the thread to be terminated
     will unlock it. As soon as we see the semaphore
     unlocked, we know that the thread has exited.
  */
  init_completion(&kthread->startstop_sem);

  /* We need to do a memory barrier here to be sure that
     the flags are visible on all CPUs. 
  */
  mb();

  /* set flag to request thread termination */
  kthread->terminate = 1;

  /* We need to do a memory barrier here to be sure that
     the flags are visible on all CPUs. 
  */
  mb();

  kill_proc(kthread->pid, SIGKILL, 1);

  /* block till thread terminated */
  wait_for_completion(&kthread->startstop_sem);

  /* release the big kernel lock */
#ifdef CONFIG_SMP
  unlock_kernel();
#endif

#if 0
  /* now we are sure the thread is in zombie state. We
     notify keventd to clean the process up.
  */
  kill_proc(2, SIGCHLD, 1);
#endif
}

/* initialize new created thread. Called by the new thread. */
void init_kthread(
    kthread_t *kthread,
    const char *name,
    int boost)
{
  /* lock the kernel. A new kernel thread starts without
     the big kernel lock, regardless of the lock state
     of the creator (the lock level is *not* inheritated)
  */
#ifdef CONFIG_SMP
  lock_kernel();
#endif

#if NEWTHREAD
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  daemonize(name);
#else
  daemonize();
  reparent_to_init();
#endif
#endif

  /* set signal mask to what we want to respond */
  siginitsetinv(&current->blocked, sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGTERM));

  spin_lock(&current->SIGMASK_LOCK);
  flush_signals(current);
  spin_unlock(&current->SIGMASK_LOCK);

  /* initialise wait queue */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  init_waitqueue_head(&kthread->queue);
#else
  init_waitqueue(&kthread->queue);
#endif

  /* initialise termination flag */
  kthread->terminate = 0;

  /* set name of this process (max 15 chars + 0 !) */
  strcpy(current->comm, name);

  if (boost) {
    /* set thread priority to real time */
    current->policy = SCHED_FIFO;
    current->rt_priority = boost;

    mb(); /* Flush memory */
  }

  /* let others run */
#ifdef CONFIG_SMP
  unlock_kernel();
#endif

  /* tell the creator that we are ready and let him continue */
  complete(&kthread->startstop_sem);
}

/* cleanup of thread. Called by the exiting thread. */
void exit_kthread(
    kthread_t *kthread)
{
  /* we are terminating */

  /* lock the kernel, the exit will unlock it */
#ifdef CONFIG_SMP
  lock_kernel();
#endif

  kthread->pid = 0;
  mb();

  /* notify the stop_kthread() routine that we are terminating. */
  complete_and_exit(&kthread->startstop_sem,0);

  /* the kernel_thread that called clone() does a do_exit here. */

  /* there is no race here between execution of the "killer" and real 
     termination
     of the thread (race window between up and do_exit), since both the
     thread and the "killer" function are running with the kernel lock held.
     The kernel lock will be freed after the thread exited, so the code
     is really not executed anymore as soon as the unload functions gets
     the kernel lock back.
     The init process may not have made the cleanup of the process here,
     but the cleanup can be done safely with the module unloaded.
  */
}
